import logging
from .models import CspReport
from .rule_evaluator import CspRuleEvaluator
from .models import CspReport, Session

REPORT_TYPE_CSP = 'csp'
REPORT_TYPE_TRIPWIRE = 'tripwire'

logger = logging.getLogger(__name__)

def raw_csp_report_to_model(raw_data, session):
    """Takes raw csp report data and session object and creates a base CspReport model.
    
    The returned model instance is not saved yet.
    """
    report = CspReport()

    report.session = session

    if 'document-url' in raw_data:
        report.document_url = raw_data['document-url']
    else:
        report.blocked_url = raw_data.get('document-uri')

    if 'blocked-url' in raw_data:
        report.blocked_url = raw_data['blocked-url']
    else:
        report.blocked_url = raw_data.get('blocked-uri')

    report.referrer = raw_data.get('referrer')
    report.effective_directive = raw_data.get('effective-directive')
    report.original_policy = raw_data.get('original-policy')
    report.source_file = raw_data.get('source-file')
    report.sample = raw_data.get('sample')
    report.disposition = raw_data.get('disposition')
    report.status_code = raw_data.get('status-code')
    report.line_number = raw_data.get('line-number')
    report.column_number = raw_data.get('column-number')

    return report

def handle_csp_report(report_data, session_id):
    """Processes incoming CSP reports."""
    if not 'csp-report' in report_data:
        return

    csp_report_raw = report_data['csp-report']
    logger.info("Received CSP report")
    logger.info("Session ID: {}".format(session_id))
    logger.info("Report Data: {}".format(csp_report_raw))

    try:
        session = Session.objects.get(id=session_id)
    except Session.DoesNotExist:
        return
    else:
        report = raw_csp_report_to_model(csp_report_raw, session)
        evaluator = CspRuleEvaluator()
        matching_rules, ignore = evaluator.evaluate_report(report)
        if not ignore:
            report.save()
            report.matching_rules.add(*matching_rules)
            logger.info("Report saved with id {}".format(report.id))
        else:
            logger.info("Report will be ignored")
    
def handle_tripwire_report(report_data, session_id):
    """Processes incoming Tripwire reports."""
    logger.info("Received Tripwire report")
    reports = []

    # retrieve corresponding session and already received reports
    try:
        session = Session.objects.get(id=session_id)
    except Session.DoesNotExist:
        return
    else:
        reports = session.cspreport_set.all()
    
    for violation in report_data:
        # check if violation should be ignored
        evaluator = CspRuleEvaluator()
        matching_rules, ignore = evaluator.evaluate_directive(violation['source'], violation['directive'])
        if ignore:
            continue
        
        # compare violation with already known reports
        unknown_violation = True
        for report in reports:
            if violation['source'] == report.blocked_url and violation['directive'] in report.effective_directive:
                unknown_violation = False
                break
        
        if unknown_violation:
            # violation detected
            logger.info("Possible injection detected. Unreported violation:")
            logger.info(violation)
            # create and save new report
            new_report = CspReport()
            new_report.session = session
            new_report.blocked_url = violation['source']
            new_report.effective_directive = violation['directive']
            new_report.document_url = violation['document']
            new_report.disposition = REPORT_TYPE_TRIPWIRE
            new_report.save()
            new_report.matching_rules.add(*matching_rules)
            logger.info("Report saved with id {}".format(new_report.id))


# Generated by Django 3.0.5 on 2020-07-03 13:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('csp_observer', '0002_auto_20200618_1232'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cspreport',
            name='disposition',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='cspreport',
            name='original_policy',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



from django.urls import path
from . import views
from .update import update_rules

urlpatterns = [
    # business logic
    path('report/<str:report_type>/<uuid:session_id>', views.report, name='report'),
    path('result/<uuid:session_id>', views.result, name='result'),
    path('resultdetail/<uuid:session_id>', views.result_detail, name='result_detail'),
    path('master/session/<uuid:session_id>', views.master_session, name='master_session'),
    path('session/<uuid:session_id>/share', views.share_session, name='share_session'),

    # admin
    path('admin/', views.admin, name='admin'),
    path('admin/update/rules/', views.admin_update_rules, name='admin_update_rules'),
    path('admin/rule/<int:rule_id>/delete', views.delete_custom_rule, name='delete_custom_rule'),
    path('chart/observed_rule_distribution', views.chart_observed_rule_distribution, name='chart_observed_rule_distribution'),
    path('chart/reports_per_day', views.chart_reports_per_day, name='chart_reports_per_day'),

    # other
    path('privacy/', views.privacy, name='privacy'),
]



# Generated by Django 3.0.5 on 2020-08-22 15:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('csp_observer', '0010_auto_20200822_1446'),
    ]

    operations = [
        migrations.RenameField(
            model_name='csprule',
            old_name='description',
            new_name='short_description',
        ),
        migrations.AddField(
            model_name='csprule',
            name='long_description',
            field=models.TextField(blank=True, null=True),
        ),
    ]



# Generated by Django 3.0.5 on 2020-08-19 19:08

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('csp_observer', '0004_csprule_is_custom_rule'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='csprule',
            name='is_custom_rule',
        ),
        migrations.AddField(
            model_name='csprule',
            name='cause',
            field=models.CharField(choices=[('EXTENSION', 'Browser Extension'), ('BROWSER', 'Web Browser'), ('MALWARE', 'Malware'), ('OTHER', 'Other')], default='OTHER', max_length=255),
        ),
    ]



import re
import logging
import json
import base64
import uuid
import asyncio
import random
from django.urls import reverse
from . import settings as app_settings
from .models import Session
from django.template.loader import render_to_string
from django.templatetags.static import static
from .report_handlers import REPORT_TYPE_CSP, REPORT_TYPE_TRIPWIRE
from .remote import create_master_session

class CspReportMiddleware:
    """The main middleware that handles all of CSP-Observer's business logic."""

    def __init__(self, get_response):
        self.super_get_response = get_response
        self.logger = logger = logging.getLogger(__name__)

        self.reporting_group_name = "csp-observer"
        self.csp_header_name = "Content-Security-Policy"
        if app_settings.REPORT_ONLY:
            self.csp_header_name = "Content-Security-Policy-Report-Only"

        # compile regexes for all enabled paths
        self.paths_regex = [re.compile("^{}$".format(p)) for p in app_settings.ENABLED_PATHS]

        nonce_temp = ''
        # generate random lower case string
        for _ in range(10):
            random_int = random.randint(97, 122)
            nonce_temp += chr(random_int)
        self.nonce = nonce_temp

    def get_response(self, request):
        response = self.super_get_response(request)
        if app_settings.IS_MASTER_COLLECTOR:
            response = self.add_cors_header(request, response)
        return response

    def anonymize_ip(self, ip_address):
        """Removes the last two octets from the ip_address."""
        return ".".join(ip_address.split(".")[:2] + 2*["0"])

    def create_session(self, request):
        """Creates a new session object, stores it in the database and returns the session id."""
        session = Session(
            user_agent=request.META["HTTP_USER_AGENT"],
            anonymized_ip=self.anonymize_ip(request.META["REMOTE_ADDR"])
        )
        session.save()
        self.logger.debug("session created: {}".format(session.id))
        return session.id

    def get_csp_policy(self, request, session_id):
        """Returns the CSP policy string based on the current settings."""
        if app_settings.REMOTE_REPORTING:
            report_uri = "{}/report/{}/{}".format(app_settings.REMOTE_CSP_OBSERVER_URL, REPORT_TYPE_CSP, session_id)
        else:
            report_uri = request.build_absolute_uri(reverse('report', args=(REPORT_TYPE_CSP, session_id, )))
        
        # set fallback reporting directive
        middleware_directives = {
            'report-uri': [report_uri],
            'script-src': ["'nonce-{}'".format(self.nonce)],
            'style-src': ["'nonce-{}'".format(self.nonce)],
            'default-src': ["'nonce-{}'".format(self.nonce)],
        }

        if app_settings.REMOTE_REPORTING:
            middleware_directives['connect-src'] = [app_settings.REMOTE_CSP_OBSERVER_URL + "/"]

        # New Reporting API stuff (not working?!):
        # https://w3c.github.io/reporting/
        # response["Reporting-Endpoints"] = '{}="{}"'.format(self.reporting_group_name, report_uri)
        report_to_group_definition = {
            "group": self.reporting_group_name,
            "max_age": 86400,
            "include_subdomains": True,
            "endpoints": [{
                "url": report_uri,
            }]
        }
        if app_settings.USE_NEW_API:
            middleware_directives['report-to'] = [self.reporting_group_name]
            response["Report-To"] = json.dumps(report_to_group_definition)

        # merge custom csp policy from settings with required middleware directives
        final_csp_policy = app_settings.CSP_POLICIES
        for directive, values in middleware_directives.items():
            if directive in final_csp_policy and directive != 'report-uri':
                final_csp_policy[directive] = set(list(final_csp_policy[directive]) + list(values))
            else:
                final_csp_policy[directive] = values
        # build csp header string
        csp_policy_string = ""
        for directive, values in final_csp_policy.items():
            csp_policy_string += "{} {}; ".format(directive, " ".join(values))

        return csp_policy_string

    def add_csp_header(self, request, response, session_id):
        """Adds the CSP header to the response."""
        policy = self.get_csp_policy(request, session_id)
        response[self.csp_header_name] = policy
        return response

    def add_cors_header(self, request, response):
        """Adds the CORS headers required for master reporting instances."""
        origins = ' '.join(app_settings.AUTHORIZED_REPORTERS)
        if 'Access-Control-Allow-Origin' in response:
            origins = response['Access-Control-Allow-Origin'] + ' ' + origins
        response['Access-Control-Allow-Origin'] = origins
        if 'Access-Control-Request-Headers' in request.headers:
            response['Access-Control-Allow-Headers'] = request.headers['Access-Control-Request-Headers']
        return response

    def add_tripwire(self, request, response, session_id):
        """Injects the tripwire javascript component into HTML response."""
        tripwire_js_path = static('csp_observer/js/tripwire.js')
        tripwire_js_uri = request.build_absolute_uri(tripwire_js_path)
        
        policy = self.get_csp_policy(request, session_id)
        policy_b64 = base64.b64encode(str.encode(policy)).decode()

        if app_settings.REMOTE_REPORTING:
            tripwire_report_uri = "{}/report/{}/{}".format(app_settings.REMOTE_CSP_OBSERVER_URL, REPORT_TYPE_TRIPWIRE, session_id)
        else:
            tripwire_report_uri = request.build_absolute_uri(reverse('report', args=(REPORT_TYPE_TRIPWIRE, session_id, )))

        script_tag_string = '<script type="text/javascript" data-session="{}" data-policy="{}" data-report-uri="{}" src="{}"></script>'.format(
            session_id, 
            policy_b64, 
            tripwire_report_uri,
            tripwire_js_uri
        )
        response.content = response.content.replace(b'</body>', str.encode(script_tag_string + '</body>'))
        return response

    def add_script_nonce(self, request, response):
        """Injects nonce attribute into all script tags in HTML response."""
        nonce_script_tag = '<script nonce="{}"'.format(self.nonce)
        response.content = response.content.replace(b'<script', str.encode(nonce_script_tag))
        return response
    
    def add_style_nonce(self, request, response):
        """Injects nonce attribute into all style and link tags in HTML response."""
        nonce_style_tag = '<style nonce="{}"'.format(self.nonce)
        response.content = response.content.replace(b'<style', str.encode(nonce_style_tag))
        nonce_link_tag = '<link nonce="{}"'.format(self.nonce)
        response.content = response.content.replace(b'<link', str.encode(nonce_link_tag))
        return response

    def add_clientui(self, request, response, session_id):
        """Injects the clientui javascript component, responsible for the popup dialog, into HTML response."""
        if app_settings.REMOTE_REPORTING:
            result_uri = "{}/result/{}".format(app_settings.REMOTE_CSP_OBSERVER_URL, session_id)
            detailpage_uri = "{}/resultdetail/{}".format(app_settings.REMOTE_CSP_OBSERVER_URL, session_id)
        else:
            result_uri = request.build_absolute_uri(reverse('result', args=(session_id, )))
            detailpage_uri = request.build_absolute_uri(reverse('result_detail', args=(session_id, )))

        clientui_html = render_to_string("client_ui/popup.html", { 
            'session_id': session_id,
            'visibility': app_settings.CLIENTUI_VISIBILITY,
            'result_uri': result_uri,
            'detailpage_uri': detailpage_uri
        })
        response.content = response.content.replace(b'</body>', str.encode(clientui_html + '</body>'))
        return response

    def __call__(self, request):
        """Checks current request path to decide if it should be observed or not."""
        # check if path in enabled paths
        for path_regex in self.paths_regex:
            if path_regex.match(request.path):
                self.logger.debug("match for path {}".format(request.path))
                if app_settings.REMOTE_REPORTING:
                    session_id = str(uuid.uuid4())
                    self.logger.debug("creating remote session {}".format(session_id))
                    asyncio.run(create_master_session(request, session_id))
                else:
                    session_id = self.create_session(request)
                request.cspo_session_id = session_id
                response = self.get_response(request)
                response = self.add_csp_header(request, response, session_id)
                response = self.add_tripwire(request, response, session_id)
                response = self.add_clientui(request, response, session_id)
                if app_settings.USE_SCRIPT_NONCE:
                    response = self.add_script_nonce(request, response)
                if app_settings.USE_STYLE_NONCE:
                    response = self.add_style_nonce(request, response)
                return response
        
        return self.get_response(request)



import urllib.parse
import urllib.request
import json
from . import settings as app_settings
from uuid import UUID

class UUIDEncoder(json.JSONEncoder):
    """https://stackoverflow.com/a/48159596"""
    def default(self, obj):
        if isinstance(obj, UUID):
            # if the obj is uuid, we simply return the value of uuid
            return obj.hex
        return json.JSONEncoder.default(self, obj)

async def create_master_session(request, session_id):
    """Creates a remote session on the master collector instance and returns status."""
    # TODO: is there a better way for programmatic path reversing on a remote host?
    remote_url = "{}/master/session/{}".format(app_settings.REMOTE_CSP_OBSERVER_URL.rstrip('/'), session_id)
    raw_data = {
        'secret': app_settings.REMOTE_SECRET,
        'user_agent': request.META["HTTP_USER_AGENT"],
        'anonymized_ip': '123'#self.anonymize_ip(request.META["REMOTE_ADDR"])
    }
    data = urllib.parse.urlencode(raw_data)
    data = data.encode('ascii')
    req = urllib.request.Request(remote_url, data)
    try:
        with urllib.request.urlopen(req) as response:
            status = response.status
            res_data = response.read()
            print(status, res_data)
    except urllib.request.URLError as e:
        print(e)

def share_session_data(session_id, data):
    """Transmits the shared session data to the voluntary reporting endpoint."""
    data = json.dumps(data, cls=UUIDEncoder).encode('ascii')
    req = urllib.request.Request(app_settings.VOLUNTARY_DATA_SHARING_URL, data)
    with urllib.request.urlopen(req) as response:
        status = response.status
        res_data = response.read()
        return (status, res_data)




import uuid
from django.db import models
from .managers import CspRuleManager

class Session(models.Model):
    """Represents a single user session (a page access)."""
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    user_agent = models.CharField(max_length=255)
    anonymized_ip = models.CharField(max_length=255)
    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return "{}@{}".format(self.id, self.created_at)

class CspRule(models.Model):
    """Represents a rule for evaluation of csp errors or whitelisting."""
    CAUSES = (
        ('extension', 'Browser Extension'),
        ('browser', 'Web Browser'),
        ('malware', 'Malware'),
        ('other', 'Other'),
    )

    global_id = models.CharField(max_length=255, blank=True, null=True)
    blocked_url = models.CharField(max_length=255, blank=True, null=True)
    effective_directive = models.CharField(max_length=255, blank=True, null=True)
    ignore = models.BooleanField(default=False)
    title = models.CharField(max_length=255, blank=True, null=True)
    short_description = models.TextField(blank=True, null=True)
    long_description = models.TextField(blank=True, null=True)
    cause = models.CharField(max_length=255, choices=CAUSES, default='other')

    objects = CspRuleManager()

    def __str__(self):
        return "{} {} ({}) {}".format(self.global_id if self.global_id else "", self.title, self.blocked_url, "[Ignore]" if self.ignore else "")

class CspReport(models.Model):
    """CSP Report model.

    Follows the new CSPViolationReportBody (https://w3c.github.io/webappsec-csp/#cspviolationreportbody)
    """
    session = models.ForeignKey(Session, to_field="id", on_delete=models.CASCADE)
    document_url = models.CharField(max_length=255)
    referrer = models.CharField(max_length=255, blank=True, null=True)
    blocked_url = models.CharField(max_length=255, blank=True, null=True)
    effective_directive = models.CharField(max_length=255)
    original_policy = models.CharField(max_length=255, blank=True, null=True)
    source_file = models.CharField(max_length=255, blank=True, null=True)
    sample = models.CharField(max_length=255, blank=True, null=True)
    disposition = models.CharField(max_length=255, blank=True, null=True)
    status_code = models.IntegerField(blank=True, null=True)
    line_number = models.IntegerField(blank=True, null=True)
    column_number = models.IntegerField(blank=True, null=True)
    matching_rules = models.ManyToManyField(CspRule, blank=True)
    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return "{}@{}".format(self.blocked_url, self.session)

class StoredConfig(models.Model):
    """Generic model for storing key/value pairs."""
    key = models.TextField(unique=True)
    value = models.TextField(blank=True, null=True)

    def __str__(self):
        return "{}={}".format(self.key, self.value)


from django.apps import AppConfig
from django.contrib.admin.apps import AdminConfig

class CspObserverConfig(AppConfig):
    name = 'csp_observer'

class CspObserverAdminConfig(AdminConfig):
    default_site = 'csp_observer.admin.CspObserverAdminSite'


# Generated by Django 3.0.5 on 2020-06-18 12:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('csp_observer', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='CspRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('blocked_url', models.CharField(blank=True, max_length=255, null=True)),
                ('effective_directive', models.CharField(blank=True, max_length=255, null=True)),
                ('ignore', models.BooleanField(default=False)),
                ('title', models.CharField(blank=True, max_length=255, null=True)),
                ('description', models.TextField(blank=True, null=True)),
            ],
        ),
        migrations.AddField(
            model_name='cspreport',
            name='matching_rules',
            field=models.ManyToManyField(blank=True, to='csp_observer.CspRule'),
        ),
    ]



from django.db import models

class CspRuleManager(models.Manager):
    """Django model manager for convenient query of global and custom rules."""
    def get_global(self):
        return super().get_queryset().exclude(global_id=None)

    def get_custom(self):
        return super().get_queryset().filter(global_id=None)


# Generated by Django 3.0.5 on 2020-05-29 08:26

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Session',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('user_agent', models.CharField(max_length=255)),
                ('anonymized_ip', models.CharField(max_length=255)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='CspReport',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('document_url', models.CharField(max_length=255)),
                ('referrer', models.CharField(blank=True, max_length=255, null=True)),
                ('blocked_url', models.CharField(blank=True, max_length=255, null=True)),
                ('effective_directive', models.CharField(max_length=255)),
                ('original_policy', models.CharField(max_length=255)),
                ('source_file', models.CharField(blank=True, max_length=255, null=True)),
                ('sample', models.CharField(blank=True, max_length=255, null=True)),
                ('disposition', models.CharField(max_length=255)),
                ('status_code', models.IntegerField(blank=True, null=True)),
                ('line_number', models.IntegerField(blank=True, null=True)),
                ('column_number', models.IntegerField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('session', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='csp_observer.Session')),
            ],
        ),
    ]



# Generated by Django 3.0.5 on 2020-08-19 19:55

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('csp_observer', '0007_config'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='Config',
            new_name='StoredConfig',
        ),
    ]



# Generated by Django 3.0.5 on 2020-08-19 18:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('csp_observer', '0003_auto_20200703_1300'),
    ]

    operations = [
        migrations.AddField(
            model_name='csprule',
            name='is_custom_rule',
            field=models.BooleanField(default=True),
        ),
    ]



# Generated by Django 3.0.5 on 2020-08-22 14:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('csp_observer', '0009_auto_20200821_0838'),
    ]

    operations = [
        migrations.AddField(
            model_name='csprule',
            name='global_id',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='storedconfig',
            name='value',
            field=models.TextField(blank=True, null=True),
        ),
    ]



# Generated by Django 3.0.5 on 2020-08-21 08:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('csp_observer', '0008_auto_20200819_1955'),
    ]

    operations = [
        migrations.AlterField(
            model_name='storedconfig',
            name='key',
            field=models.TextField(unique=True),
        ),
    ]



import re
from .models import CspRule

class CspRuleEvaluator(object):

    def __init__(self):
        self.rules = list(CspRule.objects.all())
    
    def evaluate_directive(self, url, directive):
        """Evaluates a url and directive against all rules.

        Does not differentiate between general and element directives (script-src == script-src-elem).
        Returns list of matching rules and ignore flag if the report is blacklisted.
        """
        matching_rules = []
        ignore = False
        if not url:
            return (matching_rules, ignore)
        for rule in self.rules:
            if re.match(rule.blocked_url, url):
                # url matches, now check directive if rule requires matching directive
                if (not rule.effective_directive) or (rule.effective_directive and rule.effective_directive.replace('-elem', '') == directive.replace('-elem', '')):
                    matching_rules.append(rule)
                    if rule.ignore:
                        ignore = True
        return (matching_rules, ignore)

    def evaluate_report(self, csp_report):
        """Evaluates a CSP report against all rules.

        Returns list of matching rules and ignore flag if the report is blacklisted.
        """
        return self.evaluate_directive(csp_report.blocked_url, csp_report.effective_directive)


import urllib.request
import json
import time
from . import settings as app_settings
from .models import CspRule

def retrieve_rule_data():
    """Returns the data of the global rule.json from the global data repository."""
    data = []
    with urllib.request.urlopen(app_settings.RULE_UPDATE_FILE) as f:
        raw = f.read().decode('utf-8')
        data = json.loads(raw)
    return data

def update_rules(force=False):
    """Updates the local rule database from the global data repository.

    The data is only requested if a minimum time period since the last update has passed.
    Can be forced by calling with force=True.
    """
    last_update = app_settings.get_stored(app_settings.KEY_LAST_RULE_UPDATE, default=None)
    last_update = float(last_update if last_update else 0)

    if not force:
        time_since_update = time.time() - last_update
        if time_since_update < app_settings.RULE_UPDATE_INTERVAL:
            raise Exception('Already updated within last {} hours'.format(app_settings.RULE_UPDATE_INTERVAL / 60 / 60))

    rule_database = retrieve_rule_data()
    if rule_database['last_updated'] < last_update:
        raise Exception('Local database already on latest version')
    
    # perform a little sanity check
    current_global_rule_count = CspRule.objects.get_global().count()
    allowed_difference = 0.5
    if len(rule_database['rules']) < current_global_rule_count * allowed_difference:
        # check failed
        # TODO: proper custom error handling
        raise Exception('Sanity Check Failed. Too few global rules.')

    # delete all stored rules
    CspRule.objects.get_global().delete()

    # insert the new rules into the local database
    rule_directive_counter = 0
    for rule in rule_database['rules']:
        for directive in rule['directives']:
            db_rule = CspRule(
                global_id=rule['id'],
                title=rule['title'],
                short_description=rule['short_description'],
                long_description=rule['long_description'],
                cause=rule['cause'],
                blocked_url=directive['url'],
                effective_directive=directive['directive'] if directive['directive'] != '' else None
            )
            db_rule.save()
            rule_directive_counter += 1
    
    app_settings.put_stored(app_settings.KEY_LAST_RULE_UPDATE, time.time())

    return (current_global_rule_count, len(rule_database['rules']), rule_directive_counter)


import json
import logging
import time
from datetime import datetime, timedelta
from django.utils import timezone
from django.shortcuts import render, get_object_or_404
from django.http import HttpResponse, JsonResponse
from django.views.decorators.http import require_POST
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.clickjacking import xframe_options_exempt
from django.contrib.admin.views.decorators import staff_member_required
from django.views.decorators.cache import never_cache
from django.views import generic
from django.core import serializers
from .models import CspReport, Session, CspRule
from .report_handlers import handle_csp_report, handle_tripwire_report, REPORT_TYPE_CSP, REPORT_TYPE_TRIPWIRE
from . import settings as app_settings
from .update import update_rules
from .remote import share_session_data
from django.core.paginator import Paginator
from django.views.generic import ListView
from django import forms
from django.forms.models import model_to_dict
from django.db.models import Count
import markdown as md

logger = logging.getLogger(__name__)

@csrf_exempt
def report(request, report_type, session_id):
    """Handles CSP report-uri requests."""
    if app_settings.REMOTE_REPORTING or request.method != 'POST':
        # don't do anything if remote reporting is enabled
        return HttpResponse('')

    report_str = request.body.decode('utf-8')
    report_data = json.loads(report_str)

    if report_type == REPORT_TYPE_CSP:
        handle_csp_report(report_data, session_id)
    elif report_type == REPORT_TYPE_TRIPWIRE:
        handle_tripwire_report(report_data, session_id)

    return HttpResponse('')

def result(request, session_id):
    """Handles result requests from the clientui component."""
    if app_settings.REMOTE_REPORTING:
        # don't do anything if remote reporting is enabled
        return HttpResponse('')

    session = get_object_or_404(Session, pk=session_id)
    
    # check session creation date and wait at least RESULT_WAIT_TIME seconds before returning
    current_time = timezone.now()
    min_return_time = session.created_at + timedelta(seconds=app_settings.RESULT_WAIT_TIME)
    if current_time <= min_return_time:
        time.sleep((min_return_time - current_time).seconds) 

    reports = session.cspreport_set.all()
    rules = {}
    for report in reports:
        for rule in report.matching_rules.get_global():
            if not rule.global_id in rules:
                rules[rule.global_id] = model_to_dict(rule)
        for rule in report.matching_rules.get_custom():
            if not rule.id in rules:
                rules[rule.id] = model_to_dict(rule)
    
    return JsonResponse(list(rules.values()), safe=False)

def result_detail(request, session_id):
    """Handles the request to the scan report page where the user can view their session reports."""
    if app_settings.REMOTE_REPORTING:
        # don't do anything if remote reporting is enabled
        return HttpResponse('')

    session = get_object_or_404(Session, pk=session_id)
    reports = session.cspreport_set.all()
    rules = {}
    for report in reports:
        for rule in report.matching_rules.get_global():
            if not rule.global_id in rules:
                rules[rule.global_id] = rule
        for rule in report.matching_rules.get_custom():
            if not rule.id in rules:
                rules[rule.id] = rule
    reports_without_match = [model_to_dict(x) for x in reports.filter(matching_rules=None)]

    shared_data = {
        'session': model_to_dict(session),
        'unknown_reports': reports_without_match
    }

    # prerender markdown here so we don't need a custom templatetag
    for k in rules.keys():
        rules[k].long_description = md.markdown(rules[k].long_description, extensions=['markdown.extensions.fenced_code'])

    return render(request, 'client_ui/result_detail.html', {
        'rules': rules, 
        'other_reports': reports_without_match,
        'session': session,
        'shared_data': shared_data
    })

@require_POST
@csrf_exempt
def master_session(request, session_id):
    """Handles the remote creation of a new session if instance is master collector."""
    if not app_settings.IS_MASTER_COLLECTOR:
        return HttpResponse('')
    if not app_settings.REMOTE_SECRET or len(app_settings.REMOTE_SECRET) == 0:
        return HttpResponse('')
    if not app_settings.REMOTE_SECRET == request.POST['secret']:
        return HttpResponse('')

    user_agent = request.POST['user_agent']
    anonymized_ip = request.POST['anonymized_ip']
    session = Session(
        id=session_id,
        user_agent=user_agent,
        anonymized_ip=anonymized_ip
    )
    session.save()
    logger.debug("created session {}".format(session.id))
    return session.id

    return HttpResponse('')


class CspRuleForm(forms.ModelForm):
    class Meta:
        model = CspRule
        fields = ['title', 'blocked_url', 'short_description', 'effective_directive']

@staff_member_required
@never_cache
@csrf_exempt
def admin(request):
    """Handles access of the admin dashboard."""
    cspreports = CspReport.objects.all().order_by('-created_at')
    paginator1 = Paginator(cspreports, 5)

    csprules = CspRule.objects.get_custom().order_by('title')
    paginator2 = Paginator(csprules, 5)
    
    global_database_rules = CspRule.objects.get_global().values('global_id', 'title', 'short_description').distinct().order_by('title')
    global_rules_paginator = Paginator(global_database_rules, 5)

    page_number1 = request.GET.get('page')
    page_obj1 = paginator1.get_page(page_number1)

    page_number2 = request.GET.get('page2')
    page_obj2 = paginator2.get_page(page_number2)

    global_rules_page = request.GET.get('global_rules_page')
    global_rules_page_obj = global_rules_paginator.get_page(global_rules_page)

    settings_keys = [item for item in dir(app_settings) if (not item.startswith("__") and item.isupper())]
    settings_list = [(key, getattr(app_settings, key)) for key in settings_keys]

    last_rule_update = app_settings.get_stored(app_settings.KEY_LAST_RULE_UPDATE, default=None)
    last_rule_update = datetime.fromtimestamp(float(last_rule_update if last_rule_update else 0))

    if request.method == 'POST':
        form = CspRuleForm(request.POST)
        if form.is_valid():
            form.save(commit=True)

    return render(request, 'admin/dashboard.html', {
        'cspreports': page_obj1, 
        'csprules': page_obj2, 
        'settings': settings_list,
        'global_rules': {
            'pagination': global_rules_page_obj,
            'count': global_database_rules.count(),
            'last_updated': last_rule_update
        },
    })

@require_POST
def admin_update_rules(request):
    """Handles the manual global rule update request from the admin dashboard."""
    try:
        count_pre, new_rules, count_post = update_rules(force=True)
        return JsonResponse({
            'status': 'ok',
            'message': 'Database Updated Successfully.',
            'count_pre': count_pre,
            'new_rules': new_rules,
            'count_post': count_post
        })
    except Exception as e:
        return JsonResponse({
            'status': 'error',
            'message': str(e)
        })

@require_POST
def delete_custom_rule(request, rule_id):
    """Handles the custom rule deletion request from the admin dashboard."""
    try:
        num_deleted, _ = CspRule.objects.get(id=rule_id).delete()
        return JsonResponse({
            'status': 'ok',
            'message': 'Rule {} has been removed from the database.'.format(rule_id)
        })
    except Exception as e:
        return JsonResponse({
            'status': 'error',
            'message': str(e)
        })

@require_POST
def share_session(request, session_id):
    """Handles the user's voluntary data sharing request from the result detail (scan report) page."""
    session = get_object_or_404(Session, pk=session_id)
    reports_without_match = [model_to_dict(x) for x in session.cspreport_set.filter(matching_rules=None)]

    shared_data = {
        'session': model_to_dict(session),
        'unknown_reports': reports_without_match
    }

    try:
        status, response = share_session_data(session_id, shared_data)
        if status >= 200 and status < 300:
            return JsonResponse({
                'status': 'ok',
                'message': 'The data has been submitted. Thank you!'
            })
    except Exception as e:
        print(e)
        pass

    return JsonResponse({
        'status': 'error',
        'message': 'There was an error communicating with the sharing service.'
    })

def privacy(request):
    return render(request, 'client_ui/privacy.html')

#
# Charts
#

def chart_observed_rule_distribution(request):
    """Returns the JSON data for the observed rule distribution chart on the admin dashboard."""
    counts = {}
    first_relevant_date = timezone.now() - timedelta(days=14)
    reports = CspReport.objects.filter(created_at__gt=first_relevant_date)

    for report in reports:
        raw_data = report.matching_rules.values('global_id', 'title').annotate(observed_count=Count('id'))
        for val in raw_data:
            label = "{} ({})".format(val['title'], val['global_id'])
            if not label in counts:
                counts[label] = 0
            counts[label] += val['observed_count']
    
    chart_data = {
        'labels': [x for x in counts.keys()],
        'datasets': [{
            'data': [x for x in counts.values()],
        }]
    }

    return JsonResponse(chart_data)

def chart_reports_per_day(request):
    """Returns the JSON data for the reports per day chart on the admin dashboard."""
    counts = {}
    first_relevant_date = timezone.now() - timedelta(days=14)
    reports = CspReport.objects.filter(created_at__gt=first_relevant_date).extra(select={'day': 'date( created_at )'}).values('day') \
               .annotate(count=Count('created_at'))

    chart_data = {
        'labels': [x['day'] for x in reports],
        'datasets': [{
            'label': 'Reports Per Day',
            'data': [x['count'] for x in reports],
        }]
    }

    return JsonResponse(chart_data)


from django.conf import settings
from .models import StoredConfig

NAMESPACE = getattr(settings, 'CSP_OBSERVER_NAMESPACE' , 'CSPO')
def ns_getattr(object, name, default=None):
    return getattr(settings, '_'.join([NAMESPACE, name]), default)

REPORT_ONLY = ns_getattr(settings, 'REPORT_ONLY', True)
ENABLED_PATHS = ns_getattr(settings, 'ENABLED_PATHS', [
    "/"
])

CSP_POLICIES = ns_getattr(settings, 'CSP_POLICIES', {
    'default-src': ["'self'"],
    'script-src': ["'self'"],
    'style-src': ["'self'"],
    'connect-src': ["'self'"],
    'style-src-attr': ["'unsafe-inline'"],
})

USE_NEW_API = ns_getattr(settings, 'USE_NEW_API', False)
RESULT_WAIT_TIME = ns_getattr(settings, 'RESULT_WAIT_TIME', 10)
USE_SCRIPT_NONCE = ns_getattr(settings, 'USE_SCRIPT_NONCE', True)
USE_STYLE_NONCE = ns_getattr(settings, 'USE_STYLE_NONCE', True)

SESSION_KEEP_DAYS = ns_getattr(settings, 'SESSION_KEEP_DAYS', 14)

IS_MASTER_COLLECTOR = ns_getattr(settings, 'IS_MASTER_COLLECTOR', False)
AUTHORIZED_REPORTERS = ns_getattr(settings, 'AUTHORIZED_REPORTERS', [])
REMOTE_SECRET = ns_getattr(settings, 'REMOTE_SECRET', '')
REMOTE_REPORTING = ns_getattr(settings, 'REMOTE_REPORTING', False)
REMOTE_CSP_OBSERVER_URL = ns_getattr(settings, 'REMOTE_CSP_OBSERVER_URL', "").rstrip('/')

CLIENTUI_VISIBILITY = ns_getattr(settings, 'CLIENTUI_VISIBILITY', 'always')

RULE_UPDATE_FILE = ns_getattr(settings, 'RULE_UPDATE_FILE', 'https://raw.githubusercontent.com/flxn/csp-observer-data/master/rules.json')
RULE_UPDATE_INTERVAL = ns_getattr(settings, 'RULE_UPDATE_INTERVAL', 60 * 60 * 6) # in seconds

VOLUNTARY_DATA_SHARING_URL = 'https://csp-observer-reports.flxn.de'

#
# Database-stored config values
#

KEY_LAST_RULE_UPDATE = 'LAST_RULE_UPDATE'

def get_all_stored():
    return StoredConfig.objects.all()

def delete_all_stored():
    StoredConfig.objects.all().delete()

def put_stored(key, value):
    obj, created = StoredConfig.objects.get_or_create(key=str(key))
    obj.value = value
    obj.save()

def get_stored(key, default=None):
    try:
        obj = StoredConfig.objects.get(key=key)
    except StoredConfig.DoesNotExist:
        return default
    else:
        return str(obj.value)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='django-csp-observer',
    version='1.0',
    description='A Django app that evaluates CSP reports to identify malicious activity.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/flxn/django-csp-observer",
    author='Felix Stein, Ivan Correa',
    author_email='mail@flxn.de',
    packages=['csp_observer'],
    python_requires='>=3.6',
    license='MIT License',
    install_requires=[
        'markdown',
    ],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Framework :: Django",
        "Framework :: Django :: 3.0",
        "Environment :: Web Environment",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content"
    ],
)



# Generated by Django 3.0.5 on 2020-08-19 19:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('csp_observer', '0005_auto_20200819_1908'),
    ]

    operations = [
        migrations.AlterField(
            model_name='csprule',
            name='cause',
            field=models.CharField(choices=[('extension', 'Browser Extension'), ('browser', 'Web Browser'), ('malware', 'Malware'), ('other', 'Other')], default='other', max_length=255),
        ),
    ]
